import inspect
import hyperdiv as hd
from hyperdiv.core.element import Element
from .table_utils import cell_style, header_row, qualname


def get_signature(fn):
    return f"{fn.__name__}{inspect.signature(fn)}"


def get_doc(fn):
    tag = "[public]"
    if fn.__doc__ and fn.__doc__.startswith(tag):
        return fn.__doc__.strip()[len(tag) :]
    return None


def collect_methods(klass):
    """
    Collects public methods from the class hierarchy starting at
    `klass`. Returns two dicts of

    method_name : (signature_text, description_text)

    The first dict is all methods except Element methods. The second
    is public Element methods.

    "Public" methods are methods whose docstring starts with
    `[public]`.
    """

    methods_by_class = dict()

    def collect(klass):
        if klass in methods_by_class:
            return

        methods = dict()
        for k, v in vars(klass).items():
            if callable(v) or (type(v) is property):
                doc = get_doc(v)
                if not doc:
                    continue
                if type(v) is property:
                    if v.fget and v.fset:
                        sig = "*read/write property*"
                    else:
                        sig = "*read-only property*"

                    # If both getter and setter have public docs, put
                    # them both in the description.
                    setter_doc = get_doc(v.fset)

                    if setter_doc:
                        doc = f"*Getter:* {doc}\n\n*Setter:* {setter_doc}"

                    v = v.fget
                else:
                    sig = f"`{get_signature(v)}`"

                methods[v] = (sig, doc)

        if len(methods) > 0:
            methods_by_class[klass] = methods

        for b in klass.__bases__:
            collect(b)

    normal_methods = dict()
    utility_methods = dict()

    if not inspect.isclass(klass):
        return normal_methods, utility_methods

    collect(klass)

    for klass, methods in methods_by_class.items():
        if klass == Element:
            utility_methods.update(methods)
        else:
            normal_methods.update(methods)

    return normal_methods, utility_methods


def render_methods(methods):
    for fn, (sig, doc) in methods.items():
        with hd.scope(qualname(fn)):
            with hd.table_row():
                hd.table_cell(fn.__name__, **cell_style)
                hd.table_cell(sig, **cell_style)
                hd.table_cell(doc, **cell_style)


def method_table(title, methods, expandable=False, expanded=False, show_header=False):
    """
    Renders a method table section with an optional chevron icon that
    toggles the table's visibility when clicked.
    """

    state = hd.state(expanded=expanded)

    if len(methods) > 0:
        with hd.table_row():
            if expandable or title:
                with hd.table_cell(
                    direction="horizontal",
                    colspan=4,
                    padding=4,
                    background_color=hd.theme().surface_color,
                ):
                    if expandable:
                        header = hd.text(
                            f"**{title}**",
                            (
                                hd.Icons.ChevronRightSolid
                                if not state.expanded
                                else hd.Icons.ChevronDownSolid
                            ),
                            cursor="pointer",
                        )
                        if header.clicked:
                            state.expanded = not state.expanded
                    elif title:
                        hd.text(f"**{title}**")
        if state.expanded or not expandable:
            if show_header:
                header_row("Name", "Signature", "Description")
            render_methods(methods)


def method_tables(element):
    methods, utility_methods = collect_methods(element)

    if len(methods) == 0 and len(utility_methods) == 0:
        return

    with hd.box(gap=8, font_size="small"):
        hd.text("## Methods")

        with hd.table():
            with hd.table_body():
                method_table("", methods, show_header=True)
                method_table(
                    "Common Prop Utilities",
                    utility_methods,
                    expandable=True,
                    expanded=False,
                )
