import inspect
from functools import cache
import hyperdiv as hd
from hyperdiv.core.element import Element, RemoveProp, Prop
from hyperdiv.core.context import paused_collector
from hyperdiv.elements.styled_element import StyledElement
from .table_utils import render_value, cell_style, header_row


@cache
def collect_static_props(klass, top_level_class=None):
    """Collect the static props of an element class and return a tuple of three
    dicts mapping prop names to prop instances:

    - props: the element's top-level props
    - style_props: the props inherited from StyleElement
    - box_props: the props inherited from box/container

    Note that overridden style/box props are still grouped with style
    and box props, they're not included in toplevel props. However,
    the overridden prop instance is retained.

    `top_level_class` is the top-level combo or element class we're
    collecting props for. This only matters for grouping "box props"
    for combos. If a combo returns a box, we want to group its box
    props as "Box Props", not at the top level.
    """

    # dict of (class : dict of (prop_name : prop))

    # The dict will be ordered by class hierarchy starting with the
    # superclasses, and `klass` itself at the very end.
    prop_classes = dict()
    most_recent_class = None

    if top_level_class is None:
        top_level_class = klass

    def add_class(klass):
        nonlocal most_recent_class
        prop_classes[klass] = dict()
        most_recent_class = klass

    def set_prop(prop_name, prop):
        for klass, class_props in prop_classes.items():
            if prop_name in class_props:
                class_props[prop_name] = prop
                return
        prop_classes[most_recent_class][prop_name] = prop

    def remove_prop(prop_name):
        for klass, class_props in prop_classes.items():
            if prop_name in class_props:
                class_props.pop(prop_name)
                return

    def collect_props(klass):
        """Post-order traversal makes sure that when we update the `props`
        dict, a prop in a subclass will override a prop with the
        same name in a superclass.
        """
        for b in klass.__bases__:
            if b != Element and issubclass(b, Element):
                collect_props(b)

        add_class(klass)

        for attr_name, attr in vars(klass).items():
            if isinstance(attr, RemoveProp):
                remove_prop(attr_name)
            elif isinstance(attr, Prop):
                set_prop(attr_name, attr)

    # Populate `prop_classes`
    collect_props(klass)

    # Split `prop_classes` into three dicts. Props inherited from
    # `StyledElement` will be put in `style_props`. Props inherited
    # from `box` will be put in `box_props`. All other props will be
    # lumped in `props`.
    props = dict()
    style_props = dict()
    box_props = dict()

    # Reverse `prop_classes`, so the first item is the class
    # itself. This will order the props in `props` so klass's props
    # are at the top, and props inherited from LabelElement,
    # InputElement, etc., will be toward the bottom.
    prop_classes = {k: v for k, v in reversed(prop_classes.items())}

    for the_klass, class_props in prop_classes.items():
        if the_klass == StyledElement:
            style_props.update(class_props)
        elif the_klass == hd.container and the_klass != top_level_class:
            box_props.update(class_props)
        else:
            props.update(class_props)

    return (props, style_props, box_props)


def collect_props(klass, *args, **kwargs):
    """Collect props from an element or a combo."""

    # If it's a standard element, collect the static props
    if inspect.isclass(klass) and issubclass(klass, Element):
        return collect_static_props(klass)

    # Otherwise create an instance and get the props dynamically
    props = dict()
    style_props = dict()
    box_props = dict()

    with paused_collector():
        combo_instance = klass(*args, **kwargs)

    if not hasattr(combo_instance, "_props"):
        return props, style_props, box_props

    unorganized_props = combo_instance._props

    for propname, prop in unorganized_props.items():
        the_props, the_style_props, the_box_props = collect_static_props(
            prop.parent_element.__class__, klass
        )
        if propname in the_style_props:
            style_props[propname] = prop.static_prop
        elif propname in the_box_props:
            box_props[propname] = prop.static_prop
        else:
            props[propname] = prop.static_prop

    return props, style_props, box_props


def render_props(props):
    """
    Renders a chunk of props as one table row per prop. If the prop
    subclasses OptionProp, the options are expanded inline for quick lookup.
    """
    from hyperdiv.elements.props import OptionProp

    for propname, prop in props.items():
        with hd.scope(propname):
            with hd.table_row():
                hd.table_cell(propname, **cell_style)
                proptype = type(prop).__name__
                with hd.table_cell(**cell_style, gap=8):
                    hd.text(f"[{proptype}](/proptypes#{proptype})")
                    if isinstance(prop, OptionProp):
                        options_text = "\n".join(
                            [f"* {render_value(o)}" for o in prop.options]
                        )
                        hd.text(f"Options:\n\n{options_text}")
                value = prop.value
                hd.table_cell(render_value(value), **cell_style)
                hd.table_cell("", **cell_style)


def prop_table(title, props, expandable=True, show_header=False):
    """
    Renders a prop table section with an optional chevron icon that
    toggles the table's visibility when clicked.
    """

    # Expand small tables by default
    expanded = 0 < len(props) < 5
    state = hd.state(expanded=expanded)

    if len(props) > 0:
        with hd.table_row():
            if expandable or title:
                with hd.table_cell(
                    direction="horizontal",
                    colspan=4,
                    padding=4,
                    background_color=hd.theme().surface_color,
                ):
                    if expandable:
                        header = hd.text(
                            f"**{title}**",
                            (
                                hd.Icons.ChevronRightSolid
                                if not state.expanded
                                else hd.Icons.ChevronDownSolid
                            ),
                            cursor="pointer",
                        )
                        if header.clicked:
                            state.expanded = not state.expanded
                    elif title:
                        hd.text(f"**{title}**")
        if state.expanded or not expandable:
            if show_header:
                header_row("Name", "Type", "Default", "Description")
            render_props(props)


def prop_tables(element, *args, **kwargs):
    """
    Renders the props tables. It's actually a single table with a
    "section"'s title being a row with colspan spanning the whole
    table. This way, the columns are aligned across sections.
    """

    props, style_props, box_props = collect_props(element, *args, **kwargs)

    if len(props) == 0 and len(style_props) == 0 and len(box_props) == 0:
        return

    with hd.box(gap=8, font_size="small"):
        hd.text("## Props")
        with hd.table():
            with hd.table_body():
                prop_table("", props, expandable=False, show_header=True)
                prop_table("Style Props", style_props)
                prop_table("Box Props", box_props)
